 /* Sound Creation Library*/

// information: 
// http://ccrma.stanford.edu/courses/422/projects/WaveFormat/
// http://technology.niagarac.on.ca/courses/ctec1631/WavFileFormat.html
// http://www.sonicspot.com/guide/wavefiles.html

function WAVE(init)
{

	if(this instanceof WAVE == false)
		return new WAVE(init);

	this.version = "1.1";

	this.WAVE = []; // Wave data. (numbers from -128 to 127 or -32768 to 32767)
	this.R_WAVE = []; // Unused. stereo wavs reserved for future version.
	this.L_WAVE = []; // Unused. stereo wavs reserved for future version.
	this.FILTERED = []; // Buffer for when applying filters

	
	// Integer to little endian ByteArray
	this.i2ba = function(n, len){
		var result = CreateByteArray(0);
		var b = CreateByteArray(1);
		var c;
		while( (c = n % 256) || len ){
			b[0] = c;
			result = result.concat(b);
			n >>= 8;
			--len; 
		}
		return result;
	}
	
	// bytearray memcopy2
	this.overwrite = function(data, chunk, startfrom, iam){
		if(!startfrom) startfrom = 0;
		var i = 0;
		var l = chunk.length;
		if(data.length<startfrom+chunk.length) 
			Abort(iam+":datasource too small to copy to: have:"+data.length+" need:"+(startfrom+chunk.length));
		for(;i<l;++i){
			data[i+startfrom] = chunk[i];
		}
	}
	
	// ensure value v is: l <= v <= h
	this.clamp = function(v,l,h){return v<l?l:v>h?h:v;}

	// function repeat(c,n){var m="";var i=n; while(--i>0){m+=c;}; return m;}

	this.channels = 1;
	this.samplerate = 22050; // 8000Hz - 48000Hz (44100 cd quality)
	this.bits = 8; // 8 or 16 bits samples
	
	this.RFD = CreateByteArrayFromString("RIFF0000WAVEfmt 00000000000000000000data0000");
	// size of the FORMAT chunk, always 16 (10h) for PCM
	this.overwrite(this.RFD, this.i2ba(16, 4), 16 ,"defineFORMAT");
	// Always 01 (PCM/uncompressed)
	this.overwrite(this.RFD, this.i2ba(1, 2), 20 ,"definePCM");


	this.setRIFF = function(){
		// The file begins with the ASCII characters "RIFF" identifying it as a WAV file. 
		// The next four bytes tell us the length in bytes, which is the
		// length of the entire file minus the 8 bytes for the "RIFF" and length.
		var size = this.RFD.length - 8;
		size += this.bits>8 ? 2*this.WAVE.length : this.WAVE.length;
		this.overwrite(this.RFD, this.i2ba(size,4), 4 ,"setR");
	}

	this.setFORMAT = function(){
		// Set Channel Numbers
		this.overwrite(this.RFD, this.i2ba(this.channels,   2), 22 ,"setF1");
		// Set Sample Rate
		this.overwrite(this.RFD, this.i2ba(this.samplerate, 4), 24 ,"setF2");
		// Set Bytes Per Second
		this.overwrite(this.RFD, this.i2ba(this.samplerate*this.channels*(this.bits>>3), 4), 28 ,"setF3");
		// Set Bytes per Sample
		this.overwrite(this.RFD, this.i2ba(this.channels*(this.bits>>3), 2), 32 ,"setF3");
		// Set Bits Per Sample
		this.overwrite(this.RFD, this.i2ba(this.bits, 2), 34 ,"bitspersample");
	};

	this.setDATA = function(){
		var datasize;
		if(this.channels>1)
			datasize = this.channels*(this.bits>>3) * this.R_WAVE.length;
		else
			datasize = (this.bits>>3) * this.WAVE.length;
		this.overwrite(this.RFD, this.i2ba(datasize, 4), 40 ,"setD");
		return datasize;
	};

	// Returns a bytearray containing our wave, which can be saved or given to CreateSound()
	this.get = function(){
		this.setRIFF();
		this.setFORMAT();
		var datasize = this.setDATA();
		var wav_ba =  CreateByteArray(datasize);
		
		var i;
		if(this.channels==1 && this.bits==8){
			for(i=datasize-1;i>=0;--i){
				if(isNaN(this.WAVE[i]))this.WAVE[i] = 0;
				wav_ba[i] = Math.floor(this.clamp(this.WAVE[i] + 128), 0, 255);
			}
		}else{
			Abort("Cannot handle channels="+this.channels+", bits="+this.bits+", samplerate="+this.samplerate);
		}

		if(this.channels == 2){
			// lazy... todo...
		}
		
		return this.RFD.concat(wav_ba);
	};
	
	// Set a bytearray containing a wave. not implemented yet.
	this.set = function(){
	};
	
	// save your file to the ../other/ directory
	this.save = function(filename){
		var wav = OpenRawFile(filename,true);
		wav.write(this.get());
		wav.close();
	};
	
	// Mono. volume 0-127, durations if frames, frequency, duration in secs
	this.generate = function(wavetype, frequency, volume, duration){
		var length = Math.floor(this.samplerate * (this.bits>>3) * duration);
		this.WAVE = new Array(length);
		var p = 2 * Math.PI; // phase
		var A = volume; // Amplitude
		var fn = function(n){return 0}; // silence function
		var F = frequency / this.samplerate; // Frequency

		switch(wavetype){
			case(WAVE.SINE):
				fn = Math.sin;
				break;
			case(WAVE.TRIANGLE):
				fn = function(n){return Math.asin(Math.sin(n));} 
				break;
			case(WAVE.SQUARE):
				fn = function(n){return Math.round(Math.sin(n));}
				break;
			case(WAVE.SAWTOOTH):
				fn = function(n){return 2*((n/F*0.5)-Math.floor(n/F*0.5+0.5))}
				break;
			case(WAVE.WHITENOISE):
				fn = function(n){return Math.random(n)}
				break;
			default:
				fn = wavetype;
		}
		var t; // time
		for (t = length-1;t>=0; --t){
			this.WAVE[t] = A * fn(p*F*t);
		}
 },
 
	// http://en.wikipedia.org/wiki/Low-pass_filter
	// c = real number beween 0 and 1 (1 filters nothing, 0.1 filters a lot)
	this.lowpass = function(c){
		var l = this.WAVE.length;
		for (t = 1;t<l; ++t){
			this.WAVE[t] =  this.WAVE[t-1] + c * (this.WAVE[t] - this.WAVE[t-1])
		}
	},

	// http://en.wikipedia.org/wiki/High-pass_filter
	// c = real number beween 0 and 1
	this.highpass = function(c){
		var l = this.WAVE.length;
		var p = this.WAVE[0];
		for (t = 1;t<l; ++t){
			this.WAVE[t] =  c * (this.WAVE[t-1] + this.WAVE[t] - p)
			p = this.WAVE[t];
		}     
	};

 // Process a filter function (unused)
 this.filter = function(fu){
	var l = this.WAVE.length;
	this.FILTERED = new Array(l);
	var t = 0;
	for(;t<l;t++) {
		this.FILTERED[t] = fu(t, this.WAVE[t]);
	}
	this.WAVE = this.FILTERED;
 };
 
}

WAVE.SINE = 0; // http://en.wikipedia.org/wiki/Sine_wave
WAVE.TRIANGLE = 1; // http://en.wikipedia.org/wiki/Triangle_wave
WAVE.SQUARE = 2; // http://en.wikipedia.org/wiki/Square_wave
WAVE.SAWTOOTH = 3; // http://en.wikipedia.org/wiki/Sawtooth_wave
WAVE.WHITENOISE = 4;
